************************************************************************
      PROGRAM lpvicode
* Version 2.0.2 (12/07/20)
* It computes a number of chaos indicators. (See instructions on the
* user's manual.) The indicators are:
*
*  Module1 (uses normalized and orthogonalized deviation vectors);
*      (1) LIs (Lyapunov Indicators, aka Lyapunov characteristic
*          numbers, aka Lyapunov characteristic exponents)
*  Module2 (uses normalized deviation vectors): 
*      (2) SALI (Smaller ALignment Index)
*      (3) GALIs (Generalized ALignment Indexes)
*      (4) SD & SSNs (Spectral Distance & dynamical Spectra of 
*          Stretching Numbers)
*  Module3 (uses a normalized deviation vector, plus a second orbit):
*      (5) RLI & max LI (Relative Lyapunov Indicator & maximal LI)
*  Module4 (uses an unnormalized deviation vector):
*      (6) MEGNO & SElLCE (Mean Exponential Growth factor of Nearby
*          Orbits & Slope Estimation of the largest Lyapunov
*          Characteristic Exponent)
*      (7) FLI & OFLI (Fast Lyapunov Indicator and Orthogonal FLI)
*
* The routines in this file follow this order:
* PROGRAM lpvicode
* I/O and initialization module:
*      SUBROUTINE input
*      SUBROUTINE numbor
*      SUBROUTINE condini
*      SUBROUTINE output
* Indicators module:
*      SUBROUTINE cal_lya
*      SUBROUTINE cal_sali
*      SUBROUTINE cal_gali
*      SUBROUTINE cal_ds
*      SUBROUTINE cal_rli
*      SUBROUTINE cal_megno
*      SUBROUTINE cal_fli
*      SUBROUTINE saturation
* Mathematical module:
*      SUBROUTINE renormalize
*      SUBROUTINE gram
*      SUBROUTINE unif_sph
*      SUBROUTINE bstoer
*      SUBROUTINE bsstep
*      SUBROUTINE mmid
*      SUBROUTINE rzextr
*      SUBROUTINE svd
*      FUNCTION pythag
*      SUBROUTINE quick
*      FUNCTION localizame
*      FUNCTION azar_gauss
*      FUNCTION azar
* Dynamical module:
*      SUBROUTINE acelvar
*      SUBROUTINE phys
*      FUNCTION energy
*      SUBROUTINE potinit (provided by the user in file *.pav)
*      FUNCTION pot (provided by the user in file *.pav)
*      SUBROUTINE acelera (provided by the user in file *.pav)
*      SUBROUTINE variac (provided by the user in file *.pav)
*
* The tree of call dependencies is as follows:
* PROGRAM lpvicode
*      SUBROUTINE input
*            SUBROUTINE potinit
*      SUBROUTINE numbor
*      SUBROUTINE condini
*            SUBROUTINE unif_sph
*                 FUNCTION azar_gauss
*                       FUNCTION azar
*      SUBROUTINE bstoer
*            SUBROUTINE acelvar (=derivs)
*                  SUBROUTINE acelera (from file *.pav)
*                  SUBROUTINE variac (from file *.pav)
*            SUBROUTINE bsstep
*                  SUBROUTINE mmid
*                        SUBROUTINE acelvar (=derivs)
*                              SUBROUTINE acelera (from file *.pav)
*                              SUBROUTINE variac (from file *.pav)
*                  SUBROUTINE rzextr
*      SUBROUTINE renormalize
*      SUBROUTINE cal_lya
*            SUBROUTINE gram
*            SUBROUTINE quick
*      SUBROUTINE cal_sali
*      SUBROUTINE cal_gali
*            SUBROUTINE svd
*                  FUNCTION pythag
*      SUBROUTINE cal_ds
*            FUNCTION localizame
*      SUBROUTINE cal_megno
*      SUBROUTINE cal_fli
*            SUBROUTINE acelera (from file *.pav)
*      SUBROUTINE cal_rli
*      SUBROUTINE output
*            FUNCTION energy
*                  FUNCTION pot (from file *.pav)
*            SUBROUTINE phys
*      SUBROUTINE saturation
*            SUBROUTINE numbor
************************************************************************
      INCLUDE 'LP-VIcode.par' ! parameters
      INTEGER dim
      COMMON /dimpot/dim ! dimension declared in the potential

      INTEGER imeg
      COMMON /band/imeg ! MEGNO flag

      INTEGER np,nj(-1:nind),n0j(-1:nind),kp,k,fdump,ierr,neq,ki,kf,kt,
     & ng,nor,ninit,ngal,sspt
      DOUBLE PRECISION x0(ndimf),x1b(ndimf,0:1),t,tf,paso,lcnm,t0,tf0,
     & rli,fli,ofli,megno,lyap(ndimf),slope,gali(2:ndimf),t00,
     & rnor(ndimf),rnor1b(1),sali,xva(ndimf,0:2*ndimf+2),
     & tsats,tsatg(2:ndimf),tsatf(2),tsatm
      CHARACTER inputfile*50,pref*50
      EXTERNAL acelvar

* Input parameters and names of input and output files
      CALL input(inputfile,pref,paso,t00,tf0,n0j,fdump,ninit,sspt)
      IF(t00.GT.-1d38)t0=t00 ! the same t0 for all orbits
      IF(tf0.LT.1d38)tf=tf0  ! the same tf for all orbits
* Check dimension
      IF(ndim.NE.dim)
     & STOP 'Dim of potential in *.pav <> dim in LP-VIcode.par'
* Loop over initial conditions
      OPEN(90,FILE=inputfile)
* Orbit counter
      nor=0
      DOWHILE(.TRUE.)
         nor=nor+1
* Read initial phase space position, and tf if it corresponds
         IF(t00.GT.-1d38)THEN
            IF(tf0.LT.1d38)THEN
               READ(90,*,END=1)x0
            ELSE
               READ(90,*,END=1)x0,tf
            ENDIF
         ELSE
            IF(tf0.LT.1d38)THEN
               READ(90,*,END=1)x0,t0
            ELSE
               READ(90,*,END=1)x0,t0,tf
            ENDIF
         ENDIF
* Number of steps
         np=NINT((tf-t0)/paso)
         IF(np.GT.npmax)STOP 'lpvicode: increase npmax in LP-VIcode.par'
* Set original values; actual values may have changed by saturation
         DO k=-1,nind
            nj(k)=n0j(k)
         ENDDO
         ngal=ndimf ! Start computing all GALIs
* Print evolution to screen
         IF(nj(-1).GT.0)PRINT '(a,i5,$)',' Orbit # ',nor
* Number and order of deviation vectors to be used; MEGNO flag
         CALL numbor(nj,neq,ki,kf,kt,imeg,ng)
* Initial conditions
         CALL condini(x0,nj,ki,kf,kt,ninit,xva,x1b)
* Integration loop
         DO kp=1,np
            IF(nj(-1).GT.0.AND.MOD(kp,MAX(1,np/10)).EQ.0.AND.kp.NE.np)
     &       PRINT '(i3,a,$)',100*kp/np,'%'
            t=(kp-1)*paso+t0
            CALL bstoer(t,paso,xva,ndimf*(kt+1)+ng,ierr,acelvar)
            IF(ierr.NE.0)GOTO 2
* Renormalize deviation vectors for SALI, GALI, SD, SSN and RLI
            IF(neq.GT.0)CALL renormalize(neq,xva(1,ki),rnor)
* Indicators
            IF(nj(1).GT.0)CALL cal_lya(kp,xva(1,1),t,lyap)
            IF(nj(2).GT.0)CALL cal_sali(xva(1,ki),sali)
            IF(nj(3).GT.0)CALL cal_gali(xva(1,ki),ngal,gali)
            IF(nj(4).GT.0)CALL cal_ds(np,kp,paso,rnor,nj(4),fdump,x0)
            IF(nj(6).GT.0)CALL cal_megno(kp,xva(2,kt+1),paso,megno,
     &       slope)
            IF(nj(7).GT.0)CALL cal_fli(kp,xva,xva(1,kt),t,fli,ofli)
* RLI: second orbit
            IF(nj(5).GT.0)THEN
               imeg=0
               t=(kp-1)*paso+t0
               CALL bstoer(t,paso,x1b,ndimf*2,ierr,acelvar)
               IF(ierr.NE.0)GOTO 2
               imeg=nj(6)
               CALL renormalize(1,x1b(1,1),rnor1b)
               CALL cal_rli(kp,t,rnor(1),rnor1b(1),lcnm,rli)
            ENDIF
* Saturation: do not further compute the corresponding indicator,
* except sspt = 1.
            CALL saturation(kp,t,sali,gali,fli,ofli,megno,slope,xva,nj,
     &       neq,ki,kf,kt,imeg,ng,tsats,tsatg,tsatf,tsatm,ngal,sspt,
     &       ninit)
* Output
            CALL output(nor,pref,kp,np,nj,n0j,fdump,t,x0,xva,lyap,sali,
     &       gali,lcnm,rli,megno,slope,fli,ofli,tsats,tsatg,tsatf,tsatm,
     &       ierr)
         ENDDO
* If an error occurred, leave the orbit and continue with the next one
2        IF(ierr.NE.0)CALL output(nor,pref,kp,np,nj,n0j,fdump,t,x0,xva,
     &    lyap,sali,gali,lcnm,rli,megno,slope,fli,ofli,tsats,tsatg,
     &    tsatf,tsatm,ierr)
      ENDDO
* Close files
1     CLOSE(90)
      DO k=0,nind
         IF(n0j(k).GT.0)CLOSE(20+k)
      ENDDO
      IF(n0j(0).GE.2)CLOSE(30)
      IF(n0j(4).EQ.1)CLOSE(34)
      IF(n0j(2).GT.0)CLOSE(42)
      IF(n0j(3).GT.0)CLOSE(43)
      IF(n0j(6).GT.0)CLOSE(46)
      IF(n0j(7).GT.0)CLOSE(47)
      CLOSE(91)
      END

*-----------------------------------------------------------------------
*                     I/O and initialization module
*-----------------------------------------------------------------------

************************************************************************
      SUBROUTINE input(inputfile,pref,paso,t0,tf,nj,fdump,ninit,sspt)
      INCLUDE 'LP-VIcode.par'
      INTEGER nj(-1:nind),fdump,ninit,sspt
      DOUBLE PRECISION paso,t0,tf
      CHARACTER inputfile*50,pref*50
* Input of parameters and assorted data
************************************************************************
      INTEGER k,where
      CHARACTER f*1
      OPEN(80,FILE='LP-VIcode.in')
      READ(80,*) ! (Version number)
      READ(80,*) ! Initial conditions file
      READ(80,*)inputfile
      READ(80,*) ! Prefix for output files
      READ(80,*)pref
      READ(80,*) ! Step of integration
      READ(80,*)paso
      READ(80,*) ! Initial time; final time of integration
      READ(80,*)f
      IF(f.EQ.'f'.OR.f.EQ.'F')THEN
         t0=-1d38
         BACKSPACE(80)
         READ(80,*)f,f
         IF(f.EQ.'f'.OR.f.EQ.'F')THEN
            tf=1d38
         ELSE
            BACKSPACE(80)
            READ(80,*)f,tf
         ENDIF
      ELSE
         BACKSPACE(80)
         READ(80,*)t0,f
         IF(f.EQ.'f'.OR.f.EQ.'F')THEN
            tf=1d38
         ELSE
            BACKSPACE(80)
            READ(80,*)t0,tf
         ENDIF
      ENDIF
      READ(80,*) ! Screen (0=no, 1=yes) & orbit (0=no output, 1=file)
      READ(80,*)nj(-1),nj(0)
      READ(80,*) ! Indicators: 0=no, 1=all t, 2=only last value
      READ(80,*) ! LIs, SALI, GALIs, SD, RLI & LImax, MEGNO, FLI & OFLI
      READ(80,*)(nj(k),k=1,nind)
      READ(80,*) ! Nr. of steps between outputs
      READ(80,*)fdump
      READ(80,*) ! Initial deviation vectors: 0: at random, 1: at random
                 ! but orthonormal, 2: fixed values in routine condini
      READ(80,*)ninit
      READ(80,*) ! SALI saturation with a time-dependent potential: 0: 
                 ! normal; 1: restart dev. vectors & keep integrating
      READ(80,*)sspt
      CLOSE(80)
* Set the parameters of the potential
      CALL potinit
* Check: restarting SALI is only compatible with SALI output for all t,
* no computing of GALIs, SD/SSNs nor RLI/LImax, and random initial
* deviation vectors.
      IF(sspt.EQ.1.AND.nj(2).NE.0.AND..NOT.(nj(2).EQ.1.AND.nj(3).EQ.0
     & .AND.nj(4).EQ.0.AND.nj(5).EQ.0.AND.ninit.NE.2))THEN
         IF(nj(-1).GT.0)PRINT '(a,a/a,a/a/a,a)',
     &    '*** Warning: Restarting SALI is only compatible with SALI',
     &    ' output for all t,','    with no computing GALIs, SD/SSNs',
     &    ' or RLI/LImax, and with random initial',
     &    '    deviation vectors.','    Input parameters',
     &    ' modified to satisfy these conditions.'
         nj(2)=1
         nj(3)=0
         nj(4)=0
         nj(5)=0
         ninit=0
      ENDIF
      END

************************************************************************
      SUBROUTINE numbor(nj,neq,ki,kf,kt,imeg,ng)
      INCLUDE 'LP-VIcode.par'
      INTEGER nj(-1:nind),neq,ki,kf,kt,imeg,ng
* It computes total number and order numbers for the deviation vectors
* Input:  nj(-1:nind): flags to indicators of chaos
* Output: neq: number of deviation vectors for Modules 2/3
*         ki: order number of the first dev. vector for Modules 2/3
*         kf: order number of the last dev. vector for Modules 2/3
*         kt: order number of the dev. vector for Module 4 and last
*             dev. vector to be computed
************************************************************************
      INTEGER ki0,kf0,ng0,ndimf0
      COMMON /cuales/ki0,kf0,ng0,ndimf0 ! common with subroutine bstoer

      INTEGER k1
* Initial number of deviation vectors due to Module 1 (they are 
* different from the rest because they are orthogonalized)
      k1=0
      IF(nj(1).GT.0)k1=ndimf
* Number of (normalized) deviation vectors for Modules 2/3
      neq=0 ! default
      IF(nj(5).GT.0)neq=1               ! RLI takes 1 dev. vector
      IF(nj(2).GT.0.OR.nj(4).GT.0)neq=2 ! SALI and SD take 2 dev. vec.
      IF(nj(3).GT.0)neq=ndimf           ! GALI takes ndimf dev. vectors
* ki, kf: first and last dev. vectors of SALI/GALI/SD plus RLI
      ki=k1+1
      kf=ki+neq-1
* kt: dev. vector of MEGNO/FLI *and* last dev. vector of them all
* MEGNO/FLI take 1 dev. vector, unnormalized
      kt=kf
      IF(nj(6).GT.0.OR.nj(7).GT.0)kt=kf+1 
* Set MEGNO flag
      imeg=nj(6)
      ng=MIN(2*imeg,2) ! 0 or 2: number of equations to add
* Send indices to the integration routine
      ki0=ki
      kf0=kf
      ng0=ng
      ndimf0=ndimf
      END

************************************************************************
      SUBROUTINE condini(x0,nj,ki,kf,kt,ninit,xva,x1b)
      INCLUDE 'LP-VIcode.par'
      INTEGER nj(-1:nind),ki,kf,kt,ninit
      DOUBLE PRECISION x0(ndimf),xva(ndimf,0:2*ndimf+2),
     & x1b(ndimf,0:1)
* It copies initial conditions (IC) of the phase space variables 
* and generates those of the deviation vectors, by choosing a random
* unit vector onto the ndimf-spherical surface (ninit = 0), or 
* additionally orthonormalizing them (ninit=1), or using user-provided 
* values written down in this routine (ninit =2).
* The initial deviation vectors are the same set for all the indicators.
* Also, as usual, they are loaded with their original values for all the
* orbits.
* Input:  x0(ndimf): IC of the coordinates of the phase space
*         nj(-1:nind): indicators to be computed
*         ki,kf: order numbers of the initial and final deviation
*            vectors for Modules 2/3
*         kt: order number of the deviation vector for Module 4.
*         ninit: 0=random dev. vectors, 1=idem plus orthonormalized,
*            2=written down by the user in this routine.
* Output: xva(ndimf,0:2*ndimf+2): initial conditions
*         x1b(ndimf,0:1): IC for the 2nd orbit (RLI)
************************************************************************
      LOGICAL prim
      INTEGER i,j
      DOUBLE PRECISION vec(ndimf),dv0(ndimf,ndimf)
      SAVE prim,dv0
      DATA prim/.TRUE./

* Initial deviation vectors
      IF(prim)THEN
         prim=.FALSE.
         IF(ninit.EQ.0.OR.ninit.EQ.1)THEN
* Generate initial deviation vectors at random
            DO j=1,ndimf 
               CALL unif_sph(dv0(1,j))
            ENDDO
            IF(ninit.EQ.1)CALL gram(dv0,vec,ndimf) ! orthonormalize
         ELSE
*-------------------------------------------------------------------
* Here you may input arbitrary values for the initial dev. vectors
* Please choose "2" in the corresponding parameter of "LP-VIcode.in"
* NB: dv0(1:ndimf,k) is the k-th deviation vector
* Please comment out the STOP sentence if you use this option.
* First deviation vector:
c            dv0(1,1)=...
c            dv0(2,1)=...
*            ...
* Last deviation vector for a 4D phase space:
*            ...
c            dv0(3,4)=...
c            dv0(4,4)=...
            STOP 'lpvicode: initial deviation vectors undefined.'
*-------------------------------------------------------------------
         ENDIF
      ENDIF
* IC for equations of motion
      DO i=1,ndimf 
         xva(i,0)=x0(i)
      ENDDO
* Load initial deviation vectors
* IC of variational equations: LIs
      IF(nj(1).GT.0)THEN
         DO j=1,ndimf 
            DO i=1,ndimf
               xva(i,j)=dv0(i,j)
            ENDDO
         ENDDO
      ENDIF
* IC of variational equations: SALI, GALI, SD, and 1st RLI orbit
      IF(nj(2).GT.0.OR.nj(3).GT.0.OR.nj(4).GT.0.OR.nj(5).GT.0)THEN
         DO j=ki,kf
            DO i=1,ndimf
               xva(i,j)=dv0(i,j-ki+1)
            ENDDO
         ENDDO
      ENDIF
* IC of eq. of motion and variational equations: 2nd RLI orbit
      IF(nj(5).GT.0)THEN
         DO i=1,ndimf
            x1b(i,0)=x0(i)
         ENDDO
         x1b(1,0)=x1b(1,0)+rlieps
         DO i=1,ndimf
            x1b(i,1)=dv0(i,1) ! same deviation vector as mean orbit
         ENDDO
      ENDIF
* IC of variational equations: MEGNO, FLI
      IF(nj(6).GT.0.OR.nj(7).GT.0)THEN
         DO i=1,ndimf
            xva(i,kt)=dv0(i,1)
         ENDDO
         IF(nj(6).GT.0)THEN
            xva(1,kt+1)=0d0
            xva(2,kt+1)=0d0
         ENDIF
      ENDIF
      END

************************************************************************
      SUBROUTINE output(nor,pref,kp,np,nj,n0j,fdump,t,x0,xva,lyap,sali,
     & gali,lcnm,rli,megno,slope,fli,ofli,tsats,tsatg,tsatf,tsatm,ierr)
      INCLUDE 'LP-VIcode.par'
      INTEGER kp,np,nj(-1:nind),n0j(-1:nind),fdump,nor,ierr
      DOUBLE PRECISION t,x0(ndimf),xva(ndimf),lyap(ndimf),sali,
     & gali(2:ndimf),lcnm,rli,megno,slope,fli,ofli,tsats,tsatg(2:ndimf),
     & tsatf(2),tsatm
      CHARACTER pref*50
* Output of orbit and indicators to files.
************************************************************************
      INTEGER ju,k,kploc
      CHARACTER f*2,f1*2,f2*2,f3*2,f4*2,ff*2,fffm2*2
      DOUBLE PRECISION e0,ef,energy
      SAVE e0,f,f1,f2,f3,f4,ff,fffm2

* First point of the first orbit
      IF(nor.EQ.1.AND.kp.EQ.1)THEN
* Formats
         WRITE(f,'(i2.2)')ndimf
         WRITE(f1,'(i2.2)')ndimf+1
         WRITE(f2,'(i2.2)')ndimf+2
         WRITE(f3,'(i2.2)')ndimf+3
         WRITE(f4,'(i2.2)')ndimf+4
         WRITE(ff,'(i2.2)')2*ndimf
         WRITE(fffm2,'(i2.2)')3*ndimf-2
* Open the output files
         ju=INDEX(pref,' ')-1
         IF(nj(0).GE.1)OPEN(20,FILE=pref(1:ju)//'.orb')
         IF(nj(0).GE.2)OPEN(30,FILE=pref(1:ju)//'.geo')
         IF(nj(1).GT.0)OPEN(21,FILE=pref(1:ju)//'.lyap')
         IF(nj(2).GT.0)OPEN(22,FILE=pref(1:ju)//'.sali')
         IF(nj(3).GT.0)OPEN(23,FILE=pref(1:ju)//'.gali')
         IF(nj(4).GT.0)OPEN(24,FILE=pref(1:ju)//'.sd')
         IF(nj(4).EQ.1)OPEN(34,FILE=pref(1:ju)//'.ssn')
         IF(nj(5).GT.0)OPEN(25,FILE=pref(1:ju)//'.rli')
         IF(nj(6).GT.0)OPEN(26,FILE=pref(1:ju)//'.megno')
         IF(nj(7).GT.0)OPEN(27,FILE=pref(1:ju)//'.fli')
         OPEN(91,FILE=pref(1:ju)//'.ene')
* Disable output of physical quantities if not a 2D or 3D system
         IF(nj(0).GE.2.AND.ndimf.NE.4.AND.ndimf.NE.6)THEN
            WRITE(30,*)ndimf,' degrees of freedom: output disabled.'
            nj(0)=1
         ENDIF
      ENDIF
* If an error occurred, close this orbit and continue with the next one
      IF(ierr.NE.0)THEN
* Finish output when nj=1: blank lines
         IF(nj(0).GE.1)WRITE(20,*)
         IF(nj(0).GE.2)WRITE(30,*)
         IF(nj(1).EQ.1)WRITE(21,*)
         IF(nj(2).EQ.1)WRITE(22,*)
         IF(nj(3).EQ.1)WRITE(23,*)
         IF(nj(5).EQ.1)WRITE(25,*)
         IF(nj(6).EQ.1)WRITE(26,*)
         IF(nj(7).EQ.1)WRITE(27,*)
* Finish output when nj=2: all indicators set to zero
         IF(nj(1).EQ.2)WRITE(21,'('//ff//fo//')')x0,(0d0,k=1,ndimf)
         IF(n0j(2).EQ.2)WRITE(22,'('//f2//fo//')')x0,(0d0,k=1,2)
         IF(n0j(3).EQ.2)WRITE(23,'('//fffm2//fo//')')x0,
     &    (0d0,k=1,2*ndimf-2)
         IF(nj(5).EQ.2)WRITE(25,'('//f2//fo//')')x0,(0d0,k=1,2)
         IF(n0j(6).EQ.2)WRITE(26,'('//f3//fo//')')x0,(0d0,k=1,3)
         IF(n0j(7).EQ.2)WRITE(27,'('//f4//fo//')')x0,(0d0,k=1,4)
* Send error signal to energy file and to screen
         WRITE(91,*)nor,' error: step size underflow; integrated ',
     &    100*kp/np,'%.'
         IF(nj(-1).GT.0)PRINT *,
     &    ' error: step size underflow; integrated ',100*kp/np,'%.'
         RETURN
      ENDIF

      kploc=kp
* First point of the orbit
      IF(kp.EQ.1)THEN
* Initial energy
         e0=energy(ndimf,0d0,x0)
* Dump initial conditions for the orbit and associated parameters
         IF(nj(0).GE.1)WRITE(20,'('//f1//fo//')')0d0,x0
         IF(nj(0).GE.2)CALL phys(kploc,t,x0,1)
      ENDIF

* Every point of the orbit except when a dump is due
      IF(kp.NE.1.AND.MOD(kp,fdump).NE.0.AND.kp.NE.np)THEN
         IF(nj(0).GE.2)CALL phys(kploc,t,xva,0)
      ENDIF

* Dump every fdump steps (nj=1)
      IF(MOD(kp,fdump).EQ.0.OR.kp.EQ.np)THEN
         IF(nj(0).GE.1)WRITE(20,'('//f1//fo//')')t,xva
         IF(nj(0).GE.2)CALL phys(kploc,t,xva,1)
         IF(nj(1).EQ.1)WRITE(21,'('//f1//fo//')')t,lyap
         IF(nj(2).EQ.1)WRITE(22,'(2'//fo//')')t,sali
         IF(nj(3).EQ.1)WRITE(23,'('//f//fo//')')t,gali
         IF(nj(5).EQ.1)WRITE(25,'(3'//fo//')')t,rli,lcnm
         IF(nj(6).EQ.1)WRITE(26,'(3'//fo//')')t,megno,slope
         IF(nj(7).EQ.1)WRITE(27,'(3'//fo//')')t,fli,ofli
      ENDIF

* Last point of the orbit
      IF(kp.EQ.np)THEN
* Insert a blank line between orbits (nj=1)
         IF(nj(0).GE.1)WRITE(20,*)
         IF(nj(0).GE.2)WRITE(30,*)
         IF(nj(1).EQ.1)WRITE(21,*)
         IF(nj(2).EQ.1)WRITE(22,*)
         IF(nj(3).EQ.1)WRITE(23,*)
         IF(nj(5).EQ.1)WRITE(25,*)
         IF(nj(6).EQ.1)WRITE(26,*)
         IF(nj(7).EQ.1)WRITE(27,*)
* Dump the final values (nj=2)
         IF(nj(1).EQ.2)WRITE(21,'('//ff//fo//')')x0,lyap
         IF(n0j(2).EQ.2)WRITE(22,'('//f2//fo//')')x0,sali,MIN(t,tsats)
         IF(n0j(3).EQ.2)WRITE(23,'('//fffm2//fo//')')x0,gali,
     &    (MIN(t,tsatg(k)),k=2,ndimf)
         IF(nj(5).EQ.2)WRITE(25,'('//f2//fo//')')x0,rli,lcnm
         IF(n0j(6).EQ.2)WRITE(26,'('//f3//fo//')')x0,megno,slope,
     &    MIN(t,tsatm)
         IF(n0j(7).EQ.2)WRITE(27,'('//f4//fo//')')x0,fli,ofli,
     &    (MIN(t,tsatf(k)),k=1,2)
* Final energy
         ef=energy(ndimf,0d0,xva)
         WRITE(91,*)nor,' E = ',e0,' DE = ',ABS((e0-ef)/e0)
         IF(nj(-1).GT.0)PRINT '(a,1pe8.2)',' 100% DE = ',ABS((e0-ef)/e0)
      ENDIF
      END

*-----------------------------------------------------------------------
*                          Indicators module
*-----------------------------------------------------------------------

************************************************************************
      SUBROUTINE cal_lya(kp,x,t,lyap)
      INCLUDE 'LP-VIcode.par'
      INTEGER kp
      DOUBLE PRECISION x(ndimf,ndimf),t,lyap(ndimf)
* It computes the Lyapunov indicators (Lyapunov characteristic numbers 
* or Lyapunov characteristic exponents).
* Ref.: Benettin, G., Galgani, L. & Strelcyn, J. 1976, Phys. Rev. A, 14,
*    2338-2345, and Benettin, G., Galgani, L., Giorgilli, A. & Strelcyn,
*    J. 1980, Meccanica, 15, Part I 9-20; Part II 21-30. It uses the
*    Modified Gram-Schmidt algorithm.
* Input:  kp: order number of the step of integration
*         x(ndimf,ndimf): deviation vectors
*         t= time
* Output: lyap(ndimf): LIs
************************************************************************
      INTEGER i,k
      DOUBLE PRECISION alf(ndimf),rhi(ndimf)
      SAVE alf,rhi

* Initialization
      IF(kp.EQ.1)THEN
         DO i=1,ndimf
            alf(i)=0d0   ! exponents at time t
            rhi(i)=0d0   ! sum of logarithms
         ENDDO
      ENDIF
* Orthonormalization
      CALL gram(x,alf,ndimf)
* Compute LIs
      DO k=1,ndimf
         rhi(k)=rhi(k)+LOG(alf(k))
         lyap(k)=-rhi(k)/t ! NB: minus sign only for sorting
      ENDDO
* Order LIs from largest to smallest
      CALL quick(ndimf,lyap)
      DO k=1,ndimf
         lyap(k)=-lyap(k)
      ENDDO
      END

************************************************************************
      SUBROUTINE cal_sali(x,sali)
      INCLUDE 'LP-VIcode.par'
      DOUBLE PRECISION x(ndimf,2),sali
* It computes the SALI.
* Ref.: Skokos, Ch. 2001, J. Phys. A: Math. Gen., 34, 10029-10043
* Input:  x(ndimf,2): deviation vectors
* Output: sali: SALI
************************************************************************
      INTEGER i
      DOUBLE PRECISION dm,dq
      dm=0d0
      dq=0d0
      DO i=1,ndimf
         dm=dm+(x(i,1)-x(i,2))**2
         dq=dq+(x(i,1)+x(i,2))**2
      ENDDO
      sali=MIN(SQRT(dm),SQRT(dq))
      sali=MAX(sali,satsal) ! saturation
      END

************************************************************************
      SUBROUTINE cal_gali(x,ngal,gali)
      INCLUDE 'LP-VIcode.par'
      INTEGER ngal
      DOUBLE PRECISION x(ndimf,ndimf),gali(2:ndimf)
* It computes the GALIs.
* Ref.: Skokos, Ch., Bountis, T. & Antonopoulos, Ch. 2007, Physica D, 
*    231, 30-54, and Skokos, Ch., Bountis, T. & Antonopoulos, Ch. 2008,
*    The European Physical Journal Special Topics, Vol. 165, 5-14.
* Input:  x(ndimf,ndimf): deviation vectors
* Output: gali(2:ndimf): GALI_k, k=2,...,ndimf
************************************************************************
      INTEGER j,k,i
      DOUBLE PRECISION a(ndimf,ndimf),w(ndimf),v(ndimf,ndimf)

* Compute GALI_2 through GALI_ngal
      DO k=2,ngal
* Build the matrix A**-1
         DO i=1,k
            DO j=1,ndimf
               a(j,i)=x(j,i)
            ENDDO
         ENDDO
* Singular value decomposition
         CALL svd(a,ndimf,k,ndimf,ndimf,w,v)
* Compute GALI_k
         gali(k)=0d0
         DO i=1,k
            gali(k)=gali(k)+LOG10(w(i))
* Saturation
            IF(satsal.GT.0d0)gali(k)=MAX(gali(k),LOG10(satsal))
         ENDDO
      ENDDO
* Antilogarithm
      DO k=2,ngal
         gali(k)=10d0**gali(k)
      ENDDO
      END

************************************************************************
      SUBROUTINE cal_ds(np,kp,paso,rnor,nj4,fdump,x0)
      INCLUDE 'LP-VIcode.par'
      INTEGER kp,np,nj4,fdump
      DOUBLE PRECISION paso,rnor(2),x0(ndimf)
* It computes the spectral distance SD and SSN
* Ref.: Voglis, N., Contopoulos, G. & Efthymiopoulos, C. 1999, CeMDA, 
*   73, 211, 220; Voglis, N. & Contopoulos, G. 1994, J. Phys. A: Math. 
*   Gen., 27, 4899-4909, and Contopoulos, G. & Voglis, N. 1996, CeMDA,
*   64, 1-20.
* Input: np: total number of steps of integration
*        kp: order number of the step of integration
*        paso: step of integration
*        rnor(2): norms of the deviation vectors
*        nj4: whether to dump all values or only last value
*        fdump: dump SD(t) every fdump steps
************************************************************************
      INTEGER k,j,histo1(nbin),histo2(nbin),localizame
      DOUBLE PRECISION binsize,histonorm,d,ti,d2,smin,smax,ds1(nbin),
     & ds2(nbin),rnor10,rnor20,stret1(npmax),stret2(npmax),hist(0:nbin)
      CHARACTER f1*2
      SAVE rnor10,rnor20,stret1,stret2,smax,smin

* Initialization
      IF(kp.EQ.1)THEN
         rnor10=1d0
         rnor20=1d0
         smax=-1d38
         smin=1d38
      ENDIF
* During the integration, compute stretching numbers and their extrema
      stret1(kp)=LOG(rnor(1)/rnor10)
      stret2(kp)=LOG(rnor(2)/rnor20)
      smax=MAX(smax,stret1(kp),stret2(kp))
      smin=MIN(smin,stret1(kp),stret2(kp))
      rnor10=rnor(1)
      rnor20=rnor(2)
* At the end of the integration, compute SD and SSN
      IF(kp.EQ.np)THEN
* Initialization
         binsize=(smax-smin)/nbin
         hist(0)=smin
         DO j=1,nbin
            histo1(j)=0
            histo2(j)=0
            hist(j)=smin+binsize*j
         ENDDO
         hist(nbin)=smax ! avoid accumulated error
         DO k=1,np
* Locate stretching number into the corresponding cell of the histogram
            j=localizame(hist,nbin+1,stret1(k))
            histo1(j)=histo1(j)+1
            j=localizame(hist,nbin+1,stret2(k))
            histo2(j)=histo2(j)+1
* Spectral Distance SD^2
            IF((nj4.EQ.1.AND.MOD(k,fdump).EQ.0).OR.
     &       (nj4.GT.0.AND.k.EQ.np))THEN
               histonorm=1d0/(k*binsize)
               d=0d0
               DO j=1,nbin
                  ds1(j)=histo1(j)*histonorm
                  ds2(j)=histo2(j)*histonorm
                  d=d+(ds1(j)-ds2(j))**2
               ENDDO
               d2=d*binsize
               ti=k*paso
               IF(nj4.EQ.1)THEN
                  WRITE(24,'(2'//fo//')')ti,d2
               ELSE
                  WRITE(f1,'(i2.2)')ndimf+1
                  WRITE(24,'('//f1//fo//')')x0,d2
               ENDIF
            ENDIF
         ENDDO
         IF(nj4.EQ.1)WRITE(24,*)
* Dump SSN
         IF(nj4.EQ.1)THEN            
            DO j=1,nbin
               WRITE(34,'(3'//fo//')')hist(j),ds1(j),ds2(j)
            ENDDO
            WRITE(34,*)
         ENDIF
      ENDIF
      END

************************************************************************
      SUBROUTINE cal_rli(kp,t,rnor1,rnor1b,lcnm,rli)
      INTEGER kp
      DOUBLE PRECISION t,lcnm,rli,rnor1,rnor1b
* It computes RLI and maximum LI from two orbits
* Ref.: Sndor, Z., rdi, B. & Efthymiopoulos, C. 2000, CeMDA, 78, 
*   113-123, and Sndor, Z., rdi, B., Szll, A. & Funk, B. 2004, CeMDA,
*   90, 127-138.
* Input:  kp: order number of the step of integration
*         t: time
*         rnor1: norm of dev. vector of 1st orbit
*         rnor1b: norm of dev. vector of 2nd orbit
* Output: lcnm: max LI
*         rli: RLI
************************************************************************
      DOUBLE PRECISION xliap01,xliap01b,rlitmp,lcnmb
      SAVE xliap01,xliap01b,rlitmp

* Initialization
      IF(kp.EQ.1)THEN
         xliap01=0d0
         xliap01b=0d0
         rlitmp=0d0      
      ENDIF
* LI for first orbit
      xliap01=xliap01+LOG(rnor1)
      lcnm=xliap01/t
* LI for second orbit (to compute RLI)
      xliap01b=xliap01b+LOG(rnor1b)
      lcnmb=xliap01b/t
* RLI
      rlitmp=rlitmp+ABS(lcnm-lcnmb)
      rli=rlitmp/t
      END

************************************************************************
      SUBROUTINE cal_megno(kp,xmeg,paso,megno,slope)
      INTEGER kp
      DOUBLE PRECISION paso,xmeg,megno,slope
* It computes the MEGNO and the SElLCE
* Ref.: Cincotta, P. & Sim, C. 2000, A&A, 147, 205-228, and Cincotta, 
*   P., Giordano, C. & Sim, C. 2003, Physica D, 182, 151-178.
* Input:  kp: order number of the step of integration
*         xmeg: value of MEGNO*time from the MEGNO equations
*         paso: time step
* Output: megno: value of MEGNO
*         slope: SElLCE
************************************************************************
      INCLUDE 'LP-VIcode.par'

      INTEGER k,k00,i1,kant
      DOUBLE PRECISION time,sumx,sumt1,sumt12,sumxt,megnovect(npmax)
      SAVE megnovect,kant,sumx,sumt1,sumt12,sumxt,i1

* MEGNO
      time=kp*paso
      megno=xmeg/time
      megno=MIN(megno,satmeg) ! saturation
* Initialization for the slope
      IF(kp.EQ.1)THEN
         slope=0d0
         kant=1
         i1=0
         sumx=0d0
         sumt1=0d0
         sumt12=0d0
         sumxt=0d0
      ENDIF
* Least squares fit for SElLCE
      megnovect(kp)=megno
      k00=kp-INT(0.8*kp) ! take last 80% of integration
* Substract older values
      DO k=kant,k00-1
         i1=i1-1
         time=k*paso
         sumx=sumx-megnovect(k)
         sumt1=sumt1-time
         sumt12=sumt12-time**2
         sumxt=sumxt-megnovect(k)*time
      ENDDO
* Add new value
      i1=i1+1
      time=kp*paso
      sumx=sumx+megnovect(kp)
      sumt1=sumt1+time
      sumt12=sumt12+time**2
      sumxt=sumxt+megnovect(kp)*time
* SElLCE (needs 2 or more points)
      IF(kp.GT.1)slope=
     & ABS(2d0*(i1*sumxt-sumx*sumt1)/(i1*sumt12-sumt1**2))
      kant=k00
      END

************************************************************************
      SUBROUTINE cal_fli(kp,x,xv,t,fli,ofli)
      INCLUDE 'LP-VIcode.par'
      INTEGER kp
      DOUBLE PRECISION x(ndimf),xv(ndimf),t,fli,ofli
* It computes the FLI and OFLI
* Ref.: Froeschl, Cl., Gonczi, R. & Lega, E. 1997, Planet. Space Sci., 
*   45, 881-886; Lega, E. & Froeschl, Cl. 2001, CeMDA, 81, 129-147, and
*   Fouchard, M., Lega, E., Froeschl, Ch. & Froeschl, Cl 2002, CeMDA, 
*   83, 205-222.
* Input:  kp: order number of the step of integration
*         x(ndimf): orbit
*         xv(ndimf): deviation vector
*         t: time
* Output: fli: value of the FLI
*         ofli: value of the OFLI
************************************************************************
      INTEGER i
      DOUBLE PRECISION norm2,norma2,ort,esc,dx(ndimf),acc(ndim),
     & xl(ndimf)

* Initialization
      IF(kp.EQ.1)THEN
         fli=0d0
         ofli=0d0
      ENDIF
* FLI (2001 version)
      norm2=0d0
      DO i=1,ndimf
         norm2=norm2+xv(i)**2
      ENDDO
      fli=MAX(fli,SQRT(norm2))
      fli=MIN(fli,satfli) ! saturation
* OFLI: norm of the derivatives
      DO i=1,ndim
         dx(i)=x(ndim+i)
      ENDDO
* To local variables
      DO i=1,ndimf
         xl(i)=x(i)
      ENDDO
      CALL acelera(t,xl,ndimf,acc)
      DO i=1,ndim
         dx(ndim+i)=acc(i)
      ENDDO
      norma2=0d0
      DO i=1,ndimf
         norma2=norma2+dx(i)**2
      ENDDO
* OFLI: dot product between the deviation vector and the flux
      esc=0d0 
      DO i=1,ndimf
         esc=esc+xv(i)*dx(i)
      ENDDO
      esc=esc**2/norma2
* Compute OFLI
      ort=SQRT(norm2-esc)
      ofli=MAX(ofli,ort)
      ofli=MIN(ofli,satfli) ! saturation
      END

************************************************************************
      SUBROUTINE saturation(kp,t,sali,gali,fli,ofli,megno,slope,xva,
     & nj,neq,ki,kf,kt,imeg,ng,tsats,tsatg,tsatf,tsatm,ngal,sspt,ninit)
      INCLUDE 'LP-VIcode.par'
      INTEGER kp,nj(-1:nind),neq,ki,kf,kt,imeg,ng,ngal,sspt,ninit
      DOUBLE PRECISION sali,gali(2:ndimf),fli,ofli,t,megno,slope,
     & xva(ndimf,0:2*ndimf+2),tsats,tsatg(2:ndimf),tsatf(2),tsatm
* It eliminates variables of indicators which became saturated (sspt=0).
* The input sspt = 1 means that SALI is being used with a time-dependent
* potential, and the user wants to keep integrating the orbit even when
* SALI has saturated. In this case, the deviaton vectors are restarted,
* the SALI is not eliminated, and the program keeps running.
************************************************************************
      LOGICAL ss,sg(2:ndimf),sf(2),sm
      CHARACTER f*2
      INTEGER ndel,ktf,i,j,k
      DOUBLE PRECISION vec(ndimf),dv0(ndimf,2)
      SAVE ss,sg,sf,sm,f

* Initialization for each orbit
      IF(kp.EQ.1)THEN
         WRITE(f,'(i2.2)')ndimf
         ss=.TRUE.
         DO j=2,ndimf
            sg(j)=.TRUE.
         ENDDO
         sf(1)=.TRUE.
         sf(2)=.TRUE.
         sm=.TRUE.
* Saturation times
         tsats=1d38
         DO j=2,ndimf
            tsatg(j)=1d38
         ENDDO
         tsatf(1)=1d38
         tsatf(2)=1d38
         tsatm=1d38
      ENDIF
* SALI: If it is being computed and has just saturated:
      IF(nj(2).GT.0.AND.ss.AND.sali.LE.satsal)THEN
         IF(sspt.EQ.0)THEN ! normal saturation
            ss=.FALSE.
            tsats=t
            IF(nj(2).EQ.1)WRITE(22,'(2'//fo//'/)')t,sali
            IF(nj(6).NE.0.OR.nj(7).NE.0)THEN
               ndel=0
               IF(nj(3).EQ.0.AND.nj(4).EQ.0)THEN
                  IF(nj(5).EQ.0)THEN
                     ndel=2 ! delete two equations
                  ELSE
                     ndel=1 ! delete one equation
                  ENDIF
               ENDIF
               ktf=kt
               IF(nj(6).NE.0)ktf=ktf+1 ! move MEGNO equations
               DO k=kt,ktf
                  DO j=1,ndimf
                     xva(j,k-ndel)=xva(j,k)
                  ENDDO
               ENDDO
            ENDIF
            nj(2)=0
            CALL numbor(nj,neq,ki,kf,kt,imeg,ng)
         ELSE ! keep integrating after a restart of the dev. vectors
            DO j=1,2 ! new deviation vectors at random
               CALL unif_sph(dv0(1,j))
            ENDDO
            IF(ninit.EQ.1)CALL gram(dv0,vec,ndimf) ! orthonormalize
            DO j=ki,kf ! load new ICs of variational equations
               DO i=1,ndimf
                  xva(i,j)=dv0(i,j-ki+1)
               ENDDO
            ENDDO
         ENDIF
      ENDIF
* Saturation times for GALI_k, k<>2
      DO j=3,ndimf
         IF(nj(3).GT.0.AND.sg(j).AND.gali(j).LE.satsal)THEN
            sg(j)=.FALSE.
            tsatg(j)=t
* Discontinue computation of GALI_j and above
            ngal=j-1
         ENDIF
      ENDDO
* Gali_2 determines the total saturation because it is the slower one
      IF(nj(3).GT.0.AND.sg(2).AND.gali(2).LE.satsal)THEN
         sg(2)=.FALSE.
         tsatg(2)=t
         IF(nj(3).EQ.1)WRITE(23,'('//f//fo//'/)')t,gali
         ndel=ndimf-2
         IF(nj(6).NE.0.OR.nj(7).NE.0)THEN
            IF(nj(2).EQ.0.AND.nj(4).EQ.0)THEN
               IF(nj(5).EQ.0)THEN
                  ndel=ndimf ! delete ndimf equations
               ELSE
                  ndel=ndimf-1 ! delete ndimf-1 equations
               ENDIF
            ENDIF
            ktf=kt
            IF(nj(6).NE.0)ktf=ktf+1 ! move MEGNO equations
            DO k=kt,ktf
               DO j=1,ndimf
                  xva(j,k-ndel)=xva(j,k)
               ENDDO
            ENDDO
         ENDIF
         nj(3)=0
         CALL numbor(nj,neq,ki,kf,kt,imeg,ng)
      ENDIF
* MEGNO
      IF(nj(6).GT.0.AND.sm.AND.megno.GE.satmeg)THEN
         sm=.FALSE.
         tsatm=t
         IF(nj(6).EQ.1)WRITE(26,'(3'//fo//'/)')t,megno,slope
         nj(6)=0
         CALL numbor(nj,neq,ki,kf,kt,imeg,ng)
      ENDIF
* Saturation time for FLI
      IF(nj(7).GT.0.AND.sf(1).AND.fli.GE.satfli)THEN
         sf(1)=.FALSE.
         tsatf(1)=t
      ENDIF
* OFLI determines the total saturation because it is the slower one
      IF(nj(7).GT.0.AND.sf(2).AND.ofli.GE.satfli)THEN
         sf(2)=.FALSE.
         tsatf(2)=t
         IF(nj(7).EQ.1)WRITE(27,'(3'//fo//'/)')t,fli,ofli
         nj(7)=0
         CALL numbor(nj,neq,ki,kf,kt,imeg,ng)
      ENDIF
      END

*-----------------------------------------------------------------------
*                          Mathematical module
*-----------------------------------------------------------------------

************************************************************************
      SUBROUTINE renormalize(neq,x,rnor)
      INCLUDE 'LP-VIcode.par'
      INTEGER neq
      DOUBLE PRECISION x(ndimf,neq),rnor(neq)
* Normalization of deviation vectors
* Input:  neq: number of dev. vectors to be normalized
* I/O:    x(ndimf,nva): dev. vectors
* Output: rnor(neq): norms
************************************************************************
      INTEGER i,j
      DO j=1,neq
         rnor(j)=0d0
         DO i=1,ndimf
            rnor(j)=rnor(j)+x(i,j)**2
         ENDDO
         rnor(j)=SQRT(rnor(j))
         DO i=1,ndimf
            x(i,j)=x(i,j)/rnor(j)
         ENDDO
      ENDDO
      END

************************************************************************
      SUBROUTINE gram(y,mo,n)
      INTEGER n
      DOUBLE PRECISION mo(n),y(n,n)
* Modified Gram-Schmidt orthonormalization of the matrix y(n,n).
************************************************************************
      INTEGER i,j,k
      DOUBLE PRECISION esc

* Main loop
      DO i=1,n
* Remove component in direction i
         DO j=1,i-1
* yj - <yi,yj>.yi
            esc=0d0
            DO k=1,n
               esc=esc+y(k,j)*y(k,i)
            ENDDO
            DO k=1,n
               y(k,i)=y(k,i)-y(k,j)*esc
            ENDDO
         ENDDO
* Normalization
         mo(i)=0d0
         DO k=1,n
            mo(i)=mo(i)+y(k,i)**2
         ENDDO
         mo(i)=SQRT(mo(i))
         DO k=1,n
            y(k,i)=y(k,i)/mo(i)
         ENDDO
      ENDDO
      END

************************************************************************
      SUBROUTINE unif_sph(x)
      INCLUDE 'LP-VIcode.par'
      DOUBLE PRECISION x(ndimf)
* It generates a unit ndimf-vector uniformly distributed onto the
* ndimf-spherical surface.
************************************************************************
      INTEGER i
      DOUBLE PRECISION rad,azar_gauss,gdev(ndimf)

* Generate ndimf normal deviates and compute their radius
      rad=0d0
      DO i=1,ndimf
         gdev(i)=azar_gauss(0d0,1d0)
         rad=rad+gdev(i)**2
      ENDDO
      rad=SQRT(rad)
* Generate unit vector
      DO i=1,ndimf
         x(i)=gdev(i)/rad
      ENDDO
      END

************************************************************************
      SUBROUTINE bstoer(x1,h1,ystart,nvar,ierr,derivs)
      INTEGER nvar,ierr
      DOUBLE PRECISION ystart(nvar),h1,x1
      EXTERNAL derivs
* Integration of 'nvar' differential equations with Bulirsch-Stoer
* method, from ystart(x1) until ystart(x1+h1).
* x1           -> independent variable (in / updated out)
* h1           -> step to be integrated
* ystart(nvar) -> dependent variables (in / updated out)
* nvar         -> number of equations
* ierr         -> =0 OK;
*                 =1 step size underflow (bstoer)
*                 =2 step size underflow (bsstep)
*                 =3 too many steps (bstoer)
* Based on the routine of Numerical Recipes.
************************************************************************
      INTEGER maxstp,nmax ! nmax: maximum nr. of equations (variables)
      DOUBLE PRECISION eps
      PARAMETER (maxstp=10000,nmax=100,eps=1d-13)
* Number of equations for each module
      INTEGER ki0,kf0,ng0,ndimf0
      COMMON /cuales/ki0,kf0,ng0,ndimf0

      INTEGER i,nstp
      DOUBLE PRECISION h,hdid,hnext,x,dydx(nmax),y(nmax),
     & yscal(nmax),x2

* Initialization
      h=h1
      ierr=0
      x=x1
      x2=x1+h1
      h=SIGN(h,x2-x1)
      DO i=1,nvar
         y(i)=ystart(i)
      ENDDO
      DO nstp=1,maxstp
         CALL derivs(x,y,nvar,dydx)
* Allowed error (tol=eps*yscal). NB: LIs and MEGNO have different tol's.
         DO i=1,ndimf0
            yscal(i)=MAX(ABS(y(i))+ABS(h*dydx(i)),1d0)
         ENDDO
         DO i=ndimf0+1,ndimf0+ndimf0*(ki0-1)
            yscal(i)=MAX(ABS(y(i))+ABS(h*dydx(i)),10d0)
         ENDDO
         DO i=ndimf0+ndimf0*(ki0-1)+1,ndimf0+ndimf0*kf0
            yscal(i)=MAX(ABS(y(i))+ABS(h*dydx(i)),1d0)
         ENDDO
         DO i=ndimf0+ndimf0*kf0+1,nvar
            IF(ng0.GT.0)THEN
               yscal(i)=MAX(ABS(y(i))+ABS(h*dydx(i)),10d0)
            ELSE
               yscal(i)=MAX(ABS(y(i))+ABS(h*dydx(i)),1d0)
            ENDIF
         ENDDO

         IF((x+h-x2)*(x+h-x1).GT.0d0)h=x2-x
         CALL bsstep(y,dydx,nvar,x,h,eps,yscal,hdid,hnext,derivs,ierr)
         IF(ierr.NE.0)RETURN
* If we reach h1, we leave.
         IF((x-x2)*(x2-x1).GE.0d0)THEN
            x1=x1+h1
            DO i=1,nvar
               ystart(i)=y(i)
            ENDDO
            RETURN
         ENDIF
         IF(ABS(hnext).LT.0d0)THEN ! the step didn't converge
            ierr=1
            RETURN
         ENDIF
         h=hnext
      ENDDO
      ierr=3
      END
************************************************************************
      SUBROUTINE bsstep(y,dydx,nv,x,htry,eps,yscal,hdid,hnext,derivs,
     & ierr)
      INTEGER nv,ierr
      DOUBLE PRECISION y(nv),dydx(nv),x,htry,eps,yscal(nv),hdid,hnext
      EXTERNAL derivs
* It integrates a Bulirsh-Stoer step. Slave of routine bstoer.
************************************************************************
      INTEGER nmax,kmaxx,imax
      DOUBLE PRECISION safe1,safe2,redmax,redmin,tiny,scalmx
      PARAMETER (nmax=100,kmaxx=8,imax=kmaxx+1,safe1=0.25d0,safe2=0.7d0,
     & redmax=1d-5,redmin=0.7d0,tiny=1d-30,scalmx=0.1d0)

      INTEGER i,iq,k,kk,km,kmax,kopt,nseq(imax)
      DOUBLE PRECISION eps1,epsold,errmax,fact,h,red,scale,work,wrkmin,
     & xest,xnew,a(imax),alf(kmaxx,kmaxx),err(kmaxx),yerr(nmax),
     & ysav(nmax),yseq(nmax)
      LOGICAL first,reduct
      SAVE a,alf,epsold,first,kmax,kopt,nseq,xnew
      DATA first/.TRUE./,epsold/-1d0/
      DATA nseq/2,4,6,8,10,12,14,16,18/

      IF(eps.NE.epsold)THEN
         hnext=-1d29
         xnew=-1d29
         eps1=safe1*eps
         a(1)=nseq(1)+1
         DO k=1,kmaxx
            a(k+1)=a(k)+nseq(k+1)
         ENDDO
         DO iq=2,kmaxx
            DO k=1,iq-1
               alf(k,iq)=eps1**((a(k+1)-a(iq+1))/((a(iq+1)-a(1)+1d0)*
     &          (2*k+1)))
            ENDDO
         ENDDO
         epsold=eps
         DO kopt=2,kmaxx-1
            IF(a(kopt+1).GT.a(kopt)*alf(kopt-1,kopt))GOTO 1
         ENDDO
1        kmax=kopt
      ENDIF
      h=htry
      DO i=1,nv
         ysav(i)=y(i)
      ENDDO
      IF(h.NE.hnext.OR.x.NE.xnew)THEN
         first=.TRUE.
         kopt=kmax
      ENDIF
      reduct=.FALSE.
      ierr=0
2     DO k=1,kmax
         xnew=x+h
         IF(xnew.EQ.x)THEN
            ierr=2 ! step size underflow
            RETURN
         ENDIF
         CALL mmid(ysav,dydx,nv,x,h,nseq(k),yseq,derivs)
         xest=(h/nseq(k))**2
         CALL rzextr(k,xest,yseq,y,yerr,nv)
         IF(k.NE.1)THEN
            errmax=tiny
            DO i=1,nv
               errmax=MAX(errmax,ABS(yerr(i)/yscal(i)))
            ENDDO
            errmax=errmax/eps
            km=k-1
            err(km)=(errmax/safe1)**(1d0/(2*km+1))
         ENDIF
         IF(k.NE.1.AND.(k.GE.kopt-1.OR.first))THEN
            IF(errmax.LT.1d0)GOTO 4
            IF(k.EQ.kmax.OR.k.EQ.kopt+1)THEN
               red=safe2/err(km)
               GOTO 3
            ELSEIF(k.EQ.kopt)THEN
               IF(alf(kopt-1,kopt).LT.err(km))THEN
                  red=1d0/err(km)
                  GOTO 3
               ENDIF
            ELSEIF(kopt.EQ.kmax)THEN
               IF(alf(km,kmax-1).LT.err(km))THEN
                  red=alf(km,kmax-1)*safe2/err(km)
                  GOTO 3
               ENDIF
            ELSEIF(alf(km,kopt).LT.err(km))THEN
               red=alf(km,kopt-1)/err(km)
               GOTO 3
            ENDIF
         ENDIF
      ENDDO
3     red=MIN(red,redmin)
      red=MAX(red,redmax)
      h=h*red
      reduct=.TRUE.
      GOTO 2
4     x=xnew
      hdid=h
      first=.FALSE.
      wrkmin=1d35
      DO kk=1,km
         fact=MAX(err(kk),scalmx)
         work=fact*a(kk+1)
         IF(work.LT.wrkmin)THEN
            scale=fact
            wrkmin=work
            kopt=kk+1
         ENDIF
      ENDDO
      hnext=h/scale
      IF(kopt.GE.k.AND.kopt.NE.kmax.AND..NOT.reduct)THEN
         fact=MAX(scale/alf(kopt-1,kopt),scalmx)
         IF(a(kopt+1)*fact.LE.wrkmin)THEN
            hnext=h/fact
            kopt=kopt+1
         ENDIF
      ENDIF
      END
************************************************************************
      SUBROUTINE mmid(y,dydx,nvar,xs,htot,nstep,yout,derivs)
      INTEGER nvar,nstep
      DOUBLE PRECISION y(nvar),dydx(nvar),xs,htot,yout(nvar)
      EXTERNAL derivs
* It integrates by modified midpoint method. Slave of routine bstoer.
************************************************************************
      INTEGER nmax
      PARAMETER (nmax=100)

      INTEGER i,n
      DOUBLE PRECISION h,h2,swap,x,ym(nmax),yn(nmax)

      h=htot/nstep
      DO i=1,nvar
         ym(i)=y(i)
         yn(i)=y(i)+h*dydx(i)
      ENDDO
      x=xs+h
      CALL derivs(x,yn,nvar,yout)
      h2=2d0*h
      DO n=2,nstep
         DO i=1,nvar
            swap=ym(i)+h2*yout(i)
            ym(i)=yn(i)
            yn(i)=swap
         ENDDO
         x=x+h
         CALL derivs(x,yn,nvar,yout)
      ENDDO
      DO i=1,nvar
         yout(i)=0.5d0*(ym(i)+yn(i)+h*yout(i))
      ENDDO
      END
************************************************************************
      SUBROUTINE rzextr(iest,xest,yest,yz,dy,nv)
      INTEGER iest,nv
      DOUBLE PRECISION xest,dy(nv),yest(nv),yz(nv)
* Rational function extrapolation. Slave of routine bstoer.
************************************************************************
      INTEGER imax,nmax
      PARAMETER (imax=13,nmax=100)

      INTEGER j,k
      DOUBLE PRECISION b,b1,c,ddy,v,yy,d(nmax,imax),fx(imax),x(imax)
      SAVE d,x

      x(iest)=xest
      IF(iest.EQ.1)THEN
         DO j=1,nv
            yz(j)=yest(j)
            d(j,1)=yest(j)
            dy(j)=yest(j)
         ENDDO
      ELSE
         DO k=1,iest-1
            fx(k+1)=x(iest-k)/xest
         ENDDO
         DO j=1,nv
            yy=yest(j)
            v=d(j,1)
            c=yy
            d(j,1)=yy
            DO k=2,iest
               b1=fx(k)*v
               b=b1-c
               IF(b.NE.0d0)THEN
                  b=(c-v)/b
                  ddy=c*b
                  c=b1*b
               ELSE
                  ddy=v
               ENDIF
               IF(k.NE.iest)v=d(j,k)
               d(j,k)=ddy
               yy=yy+ddy
            ENDDO
            dy(j)=ddy
            yz(j)=yy
         ENDDO
      ENDIF
      END

************************************************************************
      SUBROUTINE svd(a,m,n,mp,np,w,v)
      INTEGER m,mp,n,np
      DOUBLE PRECISION a(mp,np),v(np,np),w(np)
* Singular value decomposition of the matrix a(m,n). Based on the 
* Numerical Recipes routine.
* Singular values: vector w(n)
************************************************************************
      INTEGER nmax
      PARAMETER (nmax=500)
      INTEGER i,its,j,jj,k,l,nm
      DOUBLE PRECISION anorm,c,f,g,h,s,scale,x,y,z,rv1(nmax),pythag
      g=0d0
      scale=0d0
      anorm=0d0
      DO i=1,n
         l=i+1
         rv1(i)=scale*g
         g=0d0
         s=0d0
         scale=0d0
         IF(i.LE.m)THEN
            DO k=i,m
               scale=scale+ABS(a(k,i))
            ENDDO
            IF(scale.ne.0d0)THEN
               DO k=i,m
                  a(k,i)=a(k,i)/scale
                  s=s+a(k,i)*a(k,i)
               ENDDO
               f=a(i,i)
               g=-SIGN(SQRT(s),f)
               h=f*g-s
               a(i,i)=f-g
               DO j=l,n
                  s=0d0
                  DO k=i,m
                     s=s+a(k,i)*a(k,j)
                  ENDDO
                  f=s/h
                  DO k=i,m
                     a(k,j)=a(k,j)+f*a(k,i)
                  ENDDO
               ENDDO
               DO k=i,m
                  a(k,i)=scale*a(k,i)
               ENDDO
            ENDIF
         ENDIF
         w(i)=scale*g
         g=0d0
         s=0d0
         scale=0d0
         IF((i.LE.m).AND.(i.NE.n))THEN
            DO k=l,n
               scale=scale+ABS(a(i,k))
            ENDDO
            IF(scale.NE.0d0)THEN
               DO k=l,n
                  a(i,k)=a(i,k)/scale
                  s=s+a(i,k)*a(i,k)
               ENDDO
               f=a(i,l)
               g=-SIGN(SQRT(s),f)
               h=f*g-s
               a(i,l)=f-g
               DO k=l,n
                  rv1(k)=a(i,k)/h
               ENDDO
               DO j=l,m
                  s=0d0
                  DO k=l,n
                    s=s+a(j,k)*a(i,k)
                  ENDDO
                  DO k=l,n
                    a(j,k)=a(j,k)+s*rv1(k)
                  ENDDO
               ENDDO
               DO k=l,n
                 a(i,k)=scale*a(i,k)
               ENDDO
            ENDIF
         ENDIF
         anorm=MAX(anorm,(ABS(w(i))+ABS(rv1(i))))
      ENDDO
      DO i=n,1,-1
         IF(i.LT.n)THEN
            IF(g.NE.0d0)THEN
               DO j=l,n
                  v(j,i)=(a(i,j)/a(i,l))/g
               ENDDO
               DO j=l,n
                  s=0d0
                  DO k=l,n
                     s=s+a(i,k)*v(k,j)
                  ENDDO
                  DO k=l,n
                     v(k,j)=v(k,j)+s*v(k,i)
                  ENDDO
               ENDDO
            ENDIF
            DO j=l,n
               v(i,j)=0d0
               v(j,i)=0d0
            ENDDO
         ENDIF
         v(i,i)=1d0
         g=rv1(i)
         l=i
      ENDDO
      DO i=MIN(m,n),1,-1
         l=i+1
         g=w(i)
         DO j=l,n
            a(i,j)=0d0
         ENDDO
         IF(g.NE.0d0)THEN
            g=1d0/g
            DO j=l,n
               s=0d0
               DO k=l,m
                  s=s+a(k,i)*a(k,j)
               ENDDO
               f=(s/a(i,i))*g
               DO k=i,m
                  a(k,j)=a(k,j)+f*a(k,i)
               ENDDO
            ENDDO
            DO j=i,m
               a(j,i)=a(j,i)*g
            ENDDO
         ELSE
            DO j= i,m
               a(j,i)=0d0
            ENDDO
         endif
         a(i,i)=a(i,i)+1d0
      ENDDO
      DO k=n,1,-1
         DO its=1,30
            DO l=k,1,-1
               nm=l-1
               IF((ABS(rv1(l))+anorm).EQ.anorm)GOTO 2
               IF((ABS(w(nm))+anorm).EQ.anorm)GOTO 1
            ENDDO
1           c=0d0
            s=1d0
            DO i=l,k
               f=s*rv1(i)
               rv1(i)=c*rv1(i)
               IF((ABS(f)+anorm).EQ.anorm)GOTO 2
               g=w(i)
               h=pythag(f,g)
               w(i)=h
               h=1d0/h
               c=g*h
               s=-(f*h)
               DO j=1,m
                  y=a(j,nm)
                  z=a(j,i)
                  a(j,nm)=(y*c)+(z*s)
                  a(j,i)=-(y*s)+(z*c)
               ENDDO
            ENDDO
2           z=w(k)
            IF(l.EQ.k)THEN
               IF(z.LT.0d0)THEN
                  w(k)=-z
                  DO j=1,n
                     v(j,k)=-v(j,k)
                  ENDDO
               ENDIF
               GOTO 3
            ENDIF
            IF(its.EQ.30)STOP 'no convergence in svd'
            x=w(l)
            nm=k-1
            y=w(nm)
            g=rv1(nm)
            h=rv1(k)
            f=((y-z)*(y+z)+(g-h)*(g+h))/(2d0*h*y)
            g=pythag(f,1d0)
            f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x
            c=1d0
            s=1d0
            DO j=l,nm
               i=j+1
               g=rv1(i)
               y=w(i)
               h=s*g
               g=c*g
               z=pythag(f,h)
               rv1(j)=z
               c=f/z
               s=h/z
               f= (x*c)+(g*s)
               g=-(x*s)+(g*c)
               h=y*s
               y=y*c
               DO jj=1,n
                  x=v(jj,j)
                  z=v(jj,i)
                  v(jj,j)= (x*c)+(z*s)
                  v(jj,i)=-(x*s)+(z*c)
               ENDDO
               z=pythag(f,h)
               w(j)=z
               IF(z.NE.0d0)THEN
                  z=1d0/z
                  c=f*z
                  s=h*z
               ENDIF
               f= (c*g)+(s*y)
               x=-(s*g)+(c*y)
               DO jj=1,m
                  y=a(jj,j)
                  z=a(jj,i)
                  a(jj,j)= (y*c)+(z*s)
                  a(jj,i)=-(y*s)+(z*c)
               ENDDO
            ENDDO
            rv1(l)=0d0
            rv1(k)=f
            w(k)=x
         ENDDO
3        CONTINUE
      ENDDO
      END
************************************************************************
      FUNCTION pythag(a,b)
      DOUBLE PRECISION a,b,pythag
* It computes SQRT(a**2+b**2). Based on the Numerical Recipes routine.
************************************************************************
      DOUBLE PRECISION absa,absb
      absa=ABS(a)
      absb=ABS(b)
      IF(absa.GT.absb)THEN
         pythag=absa*SQRT(1d0+(absb/absa)**2)
      ELSE
         IF(absb.EQ.0d0)THEN
            pythag=0d0
         ELSE
            pythag=absb*SQRT(1d0+(absa/absb)**2)
         ENDIF
      ENDIF
      END

************************************************************************
      SUBROUTINE quick(n,x)
      INTEGER n
      DOUBLE PRECISION x(n)
* It sorts a vector from smallest to largest values, with the quicksort 
* algorithm described in Knuth's The Art of Computing Programming
* Input:  n - number of elements to be sorted
*	  x - vector to be sorted
* Output: x - vector already sorted
* If n is less than the actual dimension of x, the routine sorts the n
* first places; the rest is left untouched.
************************************************************************
      LOGICAL flag,fleg,flig
      INTEGER r,m,l,is,i,j,ita(2,60)
      DOUBLE PRECISION aux,aux2

      m=9
*---Quicksort algorithm-------------------------------------------------
*---Q1------------------------------------------------------------------
      l=1 
      r=n 
      is=0
*---Q2------------------------------------------------------------------
      fleg=.TRUE.
      DOWHILE(fleg)
         fleg=.FALSE.
         i=l 
         j=r+1 
         aux=x(l)
*---Q3------------------------------------------------------------------
         flag=.TRUE.
         DOWHILE(flag)
            i=i+1  
            flig=.TRUE.
            DOWHILE(i.NE.n+1.AND.flig)
               IF(x(i).LT.aux)THEN
                  i=i+1 
               ELSE
                  flig=.FALSE.
               ENDIF
            ENDDO
*---Q4------------------------------------------------------------------
            j=j-1 
            DOWHILE(x(j).GT.aux)
               j=j-1
            ENDDO
*---Q6------------------------------------------------------------------
            IF(j.GT.i)THEN 
               aux2=x(i)  
               x(i)=x(j) 
               x(j)=aux2
            ELSE
               flag=.FALSE.
            ENDIF
         ENDDO
*---Q5------------------------------------------------------------------
         x(l)=x(j) 
         x(j)=aux
*---Q7------------------------------------------------------------------
         IF(r-j.GE.j-l)THEN
            IF(j-l.GT.m)THEN 
               is=is+1 
               ita(1,is)=j+1 
               ita(2,is)=r
               r=j-1 
               fleg=.TRUE.
            ELSE 
               IF(r-j.GT.m)THEN 
                  l=j+1 
                  fleg=.TRUE.
*---Q8------------------------------------------------------------------
               ELSEIF(is.NE.0)THEN
                  l=ita(1,is) 
                  r=ita(2,is) 
                  is=is-1 
                  fleg=.TRUE.
*-----------------------------------------------------------------------
               ENDIF
            ENDIF
         ELSE
            IF(r-j.GT.m)THEN 
               is=is+1 
               ita(1,is)=l 
               ita(2,is)=j-1
               l=j+1 
               fleg=.TRUE.
            ELSE 
               IF(j-l.GT.m)THEN 
                  r=j-1 
                  fleg=.TRUE.
*---Q8------------------------------------------------------------------
               ELSEIF(is.NE.0)THEN
                  l=ita(1,is) 
                  r=ita(2,is) 
                  is=is-1 
                  fleg=.TRUE.
*-----------------------------------------------------------------------
               ENDIF
            ENDIF
         ENDIF
      ENDDO
*---Q9------------------------------------------------------------------
      DO j=2,n
         IF(x(j-1).GT.x(j))THEN
            i=j-1
            aux=x(j)
            x(i+1)=x(i)
            i=i-1
            fleg=.TRUE.
            DOWHILE(i.NE.0.AND.fleg)
               IF(aux.LT.x(i))THEN
                  x(i+1)=x(i)
                  i=i-1
               ELSE
                  fleg=.FALSE.
               ENDIF
            ENDDO
            x(i+1)=aux
         ENDIF
      ENDDO
*---End quicksort-------------------------------------------------------
      END

***********************************************************************
      FUNCTION localizame(xx,n,x)
      INTEGER localizame,n
      DOUBLE PRECISION x,xx(n)
* Bisection in a vector xx(n) to find out j so as x is in between
* xx(j) and xx(j+1). If j=0 or j=n, x is outside the table.
***********************************************************************
      INTEGER j,jl,jm,ju

      jl=0
      ju=n+1
      DOWHILE(ju-jl.GT.1)
         jm=(ju+jl)/2
         IF((xx(n).GE.xx(1)).EQV.(x.GE.xx(jm)))THEN
            jl=jm
         ELSE
            ju=jm
         ENDIF
      ENDDO
      IF(x.EQ.xx(1))THEN
         j=1
      ELSEIF(x.EQ.xx(n))THEN
         j=n-1
      ELSE
         j=jl
      ENDIF
      localizame=j
      END

**********************************************************************
      FUNCTION azar_gauss(prom,disp)
      DOUBLE PRECISION azar_gauss,disp,prom
* It generates pseudo-random numbers following a Gaussian distribution 
* with mean "prom" and dispersion "disp".
* Based on the Numerical Recipes routine.
**********************************************************************
      INTEGER iset
      DOUBLE PRECISION azar,fac,gset,rsq,v1,v2,znormal
      SAVE iset,gset
      DATA iset/0/

      IF(iset.EQ.0)THEN
1        v1=2d0*azar()-1d0
         v2=2d0*azar()-1d0
         rsq=v1**2+v2**2
         IF(rsq.GE.1d0.OR.rsq.EQ.0d0)GOTO 1
         fac=SQRT(-2d0*LOG(rsq)/rsq)
         gset=v1*fac
         znormal=v2*fac
         azar_gauss=znormal*disp+prom
         iset=1
      ELSE
         azar_gauss=gset*disp+prom
         iset=0
      ENDIF
      END

************************************************************************
      FUNCTION azar()
      DOUBLE PRECISION azar
* It generates pseudo-random numbers following a uniform istribution 
* between 0 and 1.
************************************************************************
      LOGICAL yapase
      DOUBLE PRECISION y 
      SAVE
      DATA yapase/.FALSE./

      IF(.NOT.yapase)THEN
         y=43d0
         yapase=.TRUE.
      ENDIF
      y=MOD(2188942d0*y,2147483647d0)
      azar=y/2147483647d0
      END

*-----------------------------------------------------------------------
*                           Dynamical module
*-----------------------------------------------------------------------

************************************************************************
      SUBROUTINE acelvar(t,x,n,dx)
      INTEGER n
      DOUBLE PRECISION t,x(n),dx(n)
* Accelerations and variational equations
* Input:  t: independent variable
*         n: number of dependent variables
*         x(n): dependent variables
* Output: dx(n): derivatives of x
************************************************************************
      INCLUDE 'LP-VIcode.par'
      INTEGER imeg
      COMMON /band/imeg ! MEGNO flag

      INTEGER i,nvd,j
      DOUBLE PRECISION acc(ndim),prod1,prod2

* Orbit: velocities
      DO i=1,ndim
         dx(i)=x(ndim+i)
      ENDDO
* Orbit: accelerations
      CALL acelera(t,x,ndimf,acc)
      DO i=1,ndim
         dx(ndim+i)=acc(i)
      ENDDO
* Variational equations: number of deviation vectors
      IF(imeg.NE.0)THEN ! MEGNO
         nvd=(n-2)/ndimf-1
      ELSE
         nvd=n/ndimf-1
      ENDIF
      DO j=1,nvd
* Variational equations: velocities
         DO i=1,ndim
            dx(j*ndimf+i)=x(j*ndimf+ndim+i)
         ENDDO
* Variational equations: accelerations
         CALL variac(t,x,x(j*ndimf+1),ndimf,acc)
         DO i=1,ndim
            dx(j*ndimf+ndim+i)=acc(i)
         ENDDO
      ENDDO
* Equations for MEGNO
      IF(imeg.NE.0)THEN
         prod1=0d0
         prod2=0d0
         DO i=1,ndimf
            prod1=prod1+x(nvd*ndimf+i)*dx(nvd*ndimf+i)
            prod2=prod2+x(nvd*ndimf+i)**2
         ENDDO
         dx(n-1)=t*prod1/prod2
         IF(t.EQ.0d0)THEN
            dx(n)=0d0 ! 2*dx(n-1)
         ELSE
            dx(n)=2d0*x(n-1)/t
         ENDIF
      ENDIF
      END

************************************************************************
      SUBROUTINE phys(kp,t,x,indic)
      INCLUDE 'LP-VIcode.par'
      INTEGER kp,indic
      DOUBLE PRECISION t,x(ndimf)
* It computes physical and geometrical properties of the orbit.
* Input: kp: order number of the point of the orbit
*        t: time
*        x: coordinates of the phase space
*        indic: = 0 compute only periods, no output; = 1 full
*               computation and output
* Output to file:
*  ene: energy
*  ele(3): angular momentum
*  apoc: r of last apocenter
*  perc: r of last pericenter
*  nper: number of periods since the first apocenter
************************************************************************
      INTEGER nper,k,count
      DOUBLE PRECISION energy,rv,rvant,r2,ene,ele(3),apoc,perc,xloc(6),
     & xant(ndimf)
      SAVE nper,apoc,perc,xant,rvant,count

      IF(ndimf.NE.4.AND.ndimf.NE.6)STOP'phys: not a 2D or 3D system.'
* Apocenter, pericenter and periods
      IF(kp.EQ.1)THEN
         apoc=0d0
         perc=0d0
         nper=-1
         count=0
* Initialise values of the last step
         DO k=1,ndimf
            xant(k)=x(k)
         ENDDO
         rvant=0d0
         DO k=1,ndimf/2
            rvant=rvant+x(k)*x(k+ndimf/2)
         ENDDO
      ELSE
         rv=0d0
         DO k=1,ndimf/2
            rv=rv+x(k)*x(k+ndimf/2)
         ENDDO
         IF(rv*rvant.LT.0d0)THEN
* New apocenter / pericenter
            r2=x(1)**2+x(2)**2
            IF(ndimf.EQ.6)r2=r2+x(ndimf/2)**2
            IF(rv.GT.0d0)THEN
               perc=SQRT(r2)
            ELSE
               apoc=SQRT(r2)
            ENDIF
         ENDIF
* Count periods: one period = two changes of sign of V_x
         IF(x(ndimf/2+1)*xant(ndimf/2+1).LT.0d0)THEN
* If it is the first change, set nper to 0
            IF(nper.EQ.-1)THEN
               nper=0
            ELSE
               count=count+1
               IF(count.EQ.2)THEN
                  nper=nper+1
                  count=0
               ENDIF
            ENDIF
         ENDIF
* Save values of last step
         DO k=1,ndimf
            xant(k)=x(k)
         ENDDO
         rvant=rv
      ENDIF
      IF(indic.EQ.1)THEN
* Energy
         ene=energy(ndimf,t,x)
* Angular momentum
         DO k=1,ndimf
* There are compilers that don't want an explicit x(6) when x(1:4)
            xloc(k)=x(k)
         ENDDO
         IF(ndimf.EQ.4)THEN
            ele(1)=0d0
            ele(2)=0d0
            ele(3)=xloc(1)*xloc(4)-xloc(2)*xloc(3)
         ELSE
            ele(1)=xloc(2)*xloc(6)-xloc(3)*xloc(5)
            ele(2)=xloc(3)*xloc(4)-xloc(1)*xloc(6)
            ele(3)=xloc(1)*xloc(5)-xloc(2)*xloc(4)
         ENDIF
* Output
         WRITE(30,'(7'//fo//',1x,i8)')t,ene,ele,apoc,perc,nper
      ENDIF
      END

************************************************************************
      FUNCTION energy(n,t,x)
      INTEGER n
      DOUBLE PRECISION energy,t,x(n)
* It computes the energy of the orbit
* Input: n: dimension of the phase space
*        t: time (in case the potential depends on time)
*        x: coordinates of the phase space (the potential may depend on
*           the velocities)
************************************************************************
      INTEGER i
      DOUBLE PRECISION ec,ep,pot
      ec=0d0
      DO i=n/2+1,n
         ec=ec+x(i)**2
      ENDDO
      ep=pot(t,x,n)
      energy=0.5d0*ec+ep
      END

************************************************************************
* Potential, acceleration, variational equations (user-provided)
* 2D potentials
c      INCLUDE 'binney.pav'
      INCLUDE 'henon-heiles.pav'
c      INCLUDE 'logarithmic2.pav'
* 3D potentials
c      INCLUDE 'nfwtriaxial.pav'
************************************************************************
